function kernel = getKernel(dim,basisF,dBasisF,derivative)

    
    if nargin <= 3
        kernel.derivative = 0;
    else
        kernel.derivative = derivative;
    end
    %This is a Gaussian ADR kernel
    kernel.dim = dim;
    kernel.confidenceLevel = 0.95;
    kernel.evalGram = @kernelFunctionGram;
    kernel.evalPred = @kernelFunctionPred;
    kernel.gradient = @kernelGradFunction;
    kernel.eval1 = @kernelFirstDerivativeFunction;
    kernel.eval2 = @kernelSecondDerivativeFunction;
    kernel.hessian = @kernelSDFunction;
    kernel.amplitude = 1;
    kernel.lengthscale = ones(dim,1);
    kernel.obsNoiseStDev=1;
    kernel.minLengthScale=0.05;
    kernel.maxLengthScale=2;
    
    kernel.optimiseAmplitude = (0);
    kernel.optimiseLenghtscale = (ones(dim,1));
    
    kernel.confidence_level = 0.95;
    
    kernel.rangeFraction = 0.2;
    kernel.lsMultiple = 1.5;
    kernel.maxEstimate = 0;
    kernel.estimateLStrace = @estimateLengthscaleTrace;
    kernel.estimateLSfunction = @estimateLengthscaleFunction;

    if nargin <= 1
        kernel.basisFunction = @(X)[1];
        kernel.dBasisFunction = @(X)[0];
    else
        kernel.basisFunction = basisF;
        kernel.dBasisFunction = dBasisF;
    end
    
    
    %RBF kernel, computes k(x1,x2)
    function K = kernelFunctionGram(x1,x2,kernel)
        if kernel.derivative 
           K = zeros(kernel.dim+1);
           K(1,1) = kernelFunction(x1,x2,kernel);
           for i=1:kernel.dim
               %notice: kernelFirstDerivativeFunction returns the derivative
               % w.r.t the first component, i.e. x1.
              K(i+1,1) = kernelFirstDerivativeFunction(x1,x2,i,kernel);
              K(1,i+1) = -K(i+1,1);
           end
           
           for i=1:kernel.dim
              for j=i:kernel.dim 
                  K(i,j) = -kernelSecondDerivativeFunction(x1,x2,i,j,kernel);
                  if i~=j, K(j,i) = K(i,j); end
              end
           end
           
        else
           K =  kernelFunction(x1,x2,kernel);
        end
    end    


    function K = kernelFunctionPred(x,xj,kernel)
       if kernel.derivative 
           K = zeros(1+kernel.dim,1);
           K(1) = kernelFunction(x,xj,kernel);
           for i=1:kernel.dim
              K(i+1) = -kernelFirstDerivativeFunction(x,xj,i,kernel);   
           end
       else
           K = kernelFunction(x,xj,kernel);
       end
    end    


 
    %RBF kernel, computes k(x1,x2)
    function k = kernelFunction(x1,x2,kernel)
        if  isrow(x1), x1 = x1'; end
        if  isrow(x2), x2 = x2'; end
        r = (x1 - x2)./kernel.lengthscale;
        k = kernel.amplitude*exp(-r'*r);
    end


    function k = kernelGradFunction(x1,x2,kernel)
        if  isrow(x1), x1 = x1'; end
        if  isrow(x2), x2 = x2'; end
        r = (x1 - x2)./kernel.lengthscale;
        k0 = kernel.amplitude*exp(-r'*r);
        h = 2*(x1-x2)./(kernel.lengthscale.^2);
        k = -h*k0;
    end


    function k = kernelFirstDerivativeFunction(x1,x2,i,kernel)
        if  isrow(x1), x1 = x1'; end
        if  isrow(x2), x2 = x2'; end
        if i < 0 || i > kernel.dim 
            k = NaN;
            return;
        end
        r = (x1 - x2)./kernel.lengthscale;
        k0 = kernel.amplitude*exp(-r'*r);
        h = 2*(x1(i)-x2(i))/(kernel.lengthscale(i)^2);
        k = -h*k0;
    end

    function k = kernelSDFunction(x1,x2,kernel)
        if  isrow(x1), x1 = x1'; end
        if  isrow(x2), x2 = x2'; end
        r = (x1 - x2)./kernel.lengthscale;
        k0 = kernel.amplitude*exp(-r'*r);
        h = 2*(x1-x2)./(kernel.lengthscale.^2);
        h0 = 2./(kernel.lengthscale.^2);
        H = h*h'-diag(h0);
        k = H*k0;
    end


    function k = kernelSecondDerivativeFunction(x1,x2,i,j,kernel)
            if  isrow(x1), x1 = x1'; end
            if  isrow(x2), x2 = x2'; end
            if i < 0 || i > kernel.dim || j < 0 || j > kernel.dim
                k = NaN;
                return;
            end
            r = (x1 - x2)./kernel.lengthscale;
            k0 = kernel.amplitude*exp(-r'*r);
            hi = 2*(x1(i)-x2(i))/(kernel.lengthscale(i)^2);
            hj = 2*(x1(j)-x2(j))/(kernel.lengthscale(j)^2);
            h = hi*hj;
            if i==j, h = h - 2/(kernel.lengthscale(i)^2); end
            k = h*k0;
    end


    function l = estimateLengthscaleTrace(input,obs,obs1,coord,kernel)
       %estimateLengthscaleTrace - heuristic estimate of the lengthscale
       %for trace based data. coord is the coordinate of the observable for 
       %which we want to estimate the ls.
       % obs is a  m x n x T tensor, containing the observed values.
       % obs1 is a m x d x n x T tensor, containing the derivative of obs w.r.t.
       % parameters (input dimensions).
       % m is number of observables
       % n is the number of inputs
       % d is the dimension of input space, hence the number of derivatives
       % T is the number of points in the trace
       m = size(obs1,1);
       n = size(obs1,3);
       d = size(input,2);
       T = size(obs1,4);
       l = zeros(d,1);
       %case in which we pass a n x T observation matrix (n inputs at 
       %T time points), and consequently a dim x n x T obs 1st derivative
       for i=1:d
           %proiettare input su d-1 dimensioni, escudendo i, prendere punti unici
           if (d>1)
              ind = [1:(i-1),(i+1):d];
              uinp = unique(input(:,ind),'rows');
              lc = zeros(size(uinp,1),T-1);
              for j = 1:size(uinp,1)
                  for k = 2:T %start from point 2, the first is likely to have derivatives zero
                    i1 = input(:,ind) == uinp(j,:);
                    range = kernel.rangeFraction *...
                        (max(obs(coord,i1,k)) - min(obs(coord,i1,k)));
                    k1 = obs1(coord,i,i1,k); 
                    k1 = abs(k1(:));
                    if kernel.maxEstimate
                        lc(j,k-1) = (1/kernel.lsMultiple)*max(range./k1(k1>0));
                    else
                        lc(j,k-1) = (1/kernel.lsMultiple)*mean(range./k1(k1>0));
                    end
  
                  end
              end            
           else
              lc = zeros(1,T-1);
              for k = 2:T %start from point 2, the first is likely to have derivatives zero
                range = kernel.rangeFraction *...
                    (max(obs(coord,:,k)) - min(obs(coord,:,k)));
                k1 = obs1(coord,i,:,k); 
                k1 = abs(k1(:));
                if kernel.maxEstimate
                    lc(1,k-1) = (1/kernel.lsMultiple)*max(range./k1(k1>0));
                else
                    lc(1,k-1) = (1/kernel.lsMultiple)*mean(range./k1(k1>0));
                end
              end           
           end
           if kernel.maxEstimate
                l(i) = max(lc);
           else
                l(i) = mean(lc);
           end
       end       
    end



 function l = estimateLengthscaleFunction(input,obs,obs1,coord,kernel)
       %estimateLengthscaleFunction - heuristic estimate of the lengthscale
       %for trace based data. coord is the coordinate of the observable for 
       %which we want to estimate the ls.
       % obs is a  m x n tensor, containing the observed values.
       % obs1 is a m x d x n tensor, containing the derivative of obs w.r.t.
       % parameters (input dimensions).
       % m is number of observables
       % n is the number of inputs
       % d is the dimension of input space, hence the number of derivatives
       m = size(obs1,1);
       n = size(obs1,3);
       d = size(obs1,2);
       l = zeros(d,1);
       for i=1:d
           if (d>1)
              ind = [1:(i-1),(i+1):d];
              uinp = unique(input(:,ind),'rows');
              lc = [];
              for j = 1:size(uinp,1)
                i1 = input(:,ind) == uinp(j,:);
                range = kernel.rangeFraction *...
                    (max(obs(coord,i1)) - min(obs(coord,i1)));
                k1 = obs1(coord,i,i1); 
                k1 = abs(k1(:));
                if (range > 0 & sum(k1>0) > 0)
                    if kernel.maxEstimate
                        lc = [lc, (1/kernel.lsMultiple)*max(range./k1(k1>0))];
                    else
                        lc = [lc,(1/kernel.lsMultiple)*mean(range./k1(k1>0))];
                    end
                end
              end            
           else
            range = kernel.rangeFraction *...
                (max(obs(coord,:)) - min(obs(coord,:)));
            k1 = obs1(coord,i,:); 
            k1 = abs(k1(:));
            if kernel.maxEstimate
                lc = (1/kernel.lsMultiple)*max(range./k1(k1>0));
            else
                lc = (1/kernel.lsMultiple)*mean(range./k1(k1>0));
            end
           end
           if kernel.maxEstimate
                l(i) = max(lc);
           else
                l(i) = mean(lc);
           end
       end 
 end

end

