function domain = GPUCB_getDomain( nominal_vals, coord_names )
%GPUCB_getDomain generates a standard domain structure from a vector of
%model parameters. 
%   For each parameter, with nominal value p, the function defines an
%   interval spanning two orders of magnitudes, i.e. [p/10,10*p]
%   Then, parameters are transformed taking their logarithm (?) and
%   standardised, so that the actual interval for the likelihood function 
%   ranges between [-1,1].
%   Input:
%       par  -  the vector of nominal value of parameters
%       par_names  -  a cell array with parameter names (optional)

%setting dimension
domain.dimension = length(nominal_vals);
%setting names
if (nargin > 1)
    domain.names = coord_names;
else
    domain.names = {' '};
end
%setting trueRange
if (~iscolumn(nominal_vals))
    par1 = nominal_vals';
else
    par1 = nominal_vals;
end
domain.trueRange = [par1/10, par1*10];
%setting standardised range
domain.range = [-ones(domain.dimension,1),ones(domain.dimension,1)];
domain.standardise = 1;
domain.log = ones(domain.dimension,1);
%standardisation function: converts a parameter value in trueRange 
%into the standardised range, takes a matrix of parameters (which are its
%column vectors) and a params specification.
domain.stdF = @stdF;
%inverts the standardization, from the standardized interval to the
%original one.
domain.invStdF = @invStdF;
domain.belongs = @belongsTo;




    function status = belongsTo(points,domain)
       if size(points,1) ~= domain.dimension
           status = zeros(1,size(points,2));
           return;
       end
       status = zeros(1,size(points,2));
       for j=1:size(points,2)
           status(j) = true;
           for i=1:domain.dimension
              if domain.standardise
                status(j) = status(j) & ...
                    points(i,j) <= domain.range(i,2) & ...
                    points(i,j) >= domain.range(i,1); 
              else
                status(j) = status(j) & ...
                    points(i,j) <= domain.trueRange(i,2) & ...
                    points(i,j) >= domain.trueRange(i,1);   
              end  
           end
       end
       status = logical(status); 
    end


   



    function parS = stdF(par,domain) 
       low = zeros(size(par,1),1); 
       up = zeros(size(par,1),1);
       for i = 1:size(par,1)          
            if (domain.log(i))
               low(i) = log(domain.trueRange(i,1));
               up(i) = log(domain.trueRange(i,2));
            else
               low(i) = domain.trueRange(i,1);
               up(i) = domain.trueRange(i,2);
            end
       end 
       mean = (up + low)/2;
       width = up - mean;
       stdMean = (domain.range(:,1) + domain.range(:,2))/2;
       stdWidth = domain.range(:,2) - stdMean;
       parS = zeros(size(par));
       for i = 1:size(par,2) 
           p = domain.log .* log(par(:,i)) +...
               (1-domain.log) .* par(:,i);
           parS(:,i) = (((p-mean)./width).*stdWidth)+stdMean;
       end   
    end


    function parI = invStdF(par,domain)
       low = zeros(size(par,1),1); 
       up = zeros(size(par,1),1);
       for i = 1:size(par,1)          
            if (domain.log(i))
               low(i) = log(domain.trueRange(i,1));
               up(i) = log(domain.trueRange(i,2));
            else
               low(i) = domain.trueRange(i,1);
               up(i) = domain.trueRange(i,2);
            end
       end
       mean = (up + low)/2;
       width = up - mean;
       stdMean = (domain.range(:,1) + domain.range(:,2))/2;
       stdWidth = domain.range(:,2) - stdMean;
       parI = zeros(size(par));  
       for i = 1:size(par,2)      
           p = (((par(:,i)-stdMean)./stdWidth).*width)+mean;
           parI(:,i) = domain.log .* exp(p) +...
               (1-domain.log) .* p;
       end    
    end


end

