function rm = GPregressionMean(input,obs,noise,kernel)
%GPregression Computes GP regression model, contains a function for
%prediction and a function 

rm.input = input;


rm.dim = size(input,2);
rm.n = size(input,1);
rm.q = length(kernel.basisFunction(input(1,:)));

%d is the alpha-parameter of the inverse-normal gamma distribution, and
%should be set ot zero to get the uninformative prior with infinite variance 
%on sigma2.
rm.d = 0;


%check data consistency
if (size(obs,2) ~= rm.n)
    disp('input data and observations have different dimesions');
    return;
end
if (kernel.dim ~= rm.dim)
    disp('kernel has the wrong dimensions');
    return;
end

if length(noise) == 1
    rm.noise = noise*ones(rm.n,1);
else
    rm.noise = noise;
end

%Process input, computing the necessary matrices
%compute gram matrix and H matrix
rm.kernel = kernel;
rm.basisFunction = kernel.basisFunction;
rm.dBasisFunction = kernel.dBasisFunction;

rm.selfvariance = kernel.amplitude;
rm.K = zeros(rm.n);
rm.H = zeros(rm.n,rm.q);
for i = 1:rm.n
    rm.H(i,:) = kernel.basisFunction(input(i,:));
    for j = 1:rm.n
        rm.K(i,j) = kernel.evalGram(input(i,:),input(j,:),kernel);
    end
end
%forcing the matrix to be positive definite by adding a small noise term
rm.K = rm.K + diag(rm.noise.^2);
rm.L = chol(rm.K,'lower');
rm.Kinv = inv(rm.L')*inv(rm.L);
%rm.KinvH = L'\(L\H);
rm.KinvH = rm.Kinv*rm.H;
V_star_inv = rm.H'*rm.KinvH; 
Lstar = chol(V_star_inv,'lower');
rm.V_star = inv(Lstar')*inv(Lstar);
rm.conf = tinv((1+kernel.confidence_level)/2,rm.n+rm.d);

if isrow(obs)
    rm.y = obs';
else
    rm.y = obs;
end

%Kobs = L'\(L\obs);
rm.Kobs = rm.Kinv*rm.y;
%beta_hat = Lstar'\(Lstar\(H'*Kobs));
rm.beta_hat = rm.V_star*(rm.H'*rm.Kobs);
sigma2_hat = (rm.y'*rm.Kobs - rm.beta_hat'*V_star_inv*rm.beta_hat)/(rm.n+rm.d-2);
if sigma2_hat < 0
    sigma2_hat = 1e-16;
end
rm.sigma_hat = sqrt(sigma2_hat);
%rm.sigma_hat = sqrt(diag((rm.y'*rm.Kobs - rm.beta_hat'*V_star_inv*rm.beta_hat))'/(rm.n+rm.d-2));
%Kobs1 = L'\(L\(obs-H*beta_hat));
rm.Kobs1 = rm.Kinv*(rm.y-rm.H*rm.beta_hat);


rm.predict = @predict;
rm.predictSet = @predictSet;
rm.predictGradient = @predictGradient;


    function [p,std,l,u] = predict(x,rm) 
        k = zeros(rm.n,1);
        for jj=1:rm.n
           k(jj) = rm.kernel.evalPred(x,rm.input(jj,:),rm.kernel);
        end
        s = rm.selfvariance; 
        hx = rm.basisFunction(x);

        m_star = hx*rm.beta_hat + k'*rm.Kobs1;
        %Kinvk = L'\(L\k);
        Kinvk = rm.Kinv*k;

        hx1 = hx - k'*rm.KinvH;
        c_star = s - k'*Kinvk + hx1*rm.V_star*hx1';

       if c_star < 0
           c_star = abs(c_star);
       end

       std = rm.sigma_hat*sqrt(c_star);
       pm = rm.conf*std;
       p = m_star; 
       u = m_star + pm;
       l = m_star - pm;
    end


    function [m,std] = predictSet(xs,rm)
       n = size(xs,1);
       m = zeros(n,1);
       std = zeros(n,1);
       for h=1:n
           [m(h),std(h)] = rm.predict(xs(h,:),rm);
       end
    end



    function [p,std,dp,dstd] = predictGradient(x,rm)
       k = zeros(rm.n,1);
       dk = zeros(rm.n,rm.dim);
       for jj=1:rm.n
         k(jj) = rm.kernel.evalPred(x,rm.input(jj,:),rm.kernel);
         dk(jj,:) = rm.kernel.gradient(x,rm.input(jj,:),rm.kernel)';
       end
       s = rm.selfvariance; 
       hx = rm.basisFunction(x);
       dhx = rm.dBasisFunction(x);

       p =   hx*rm.beta_hat +  k'*rm.Kobs1;
       dp = dhx*rm.beta_hat + dk'*rm.Kobs1;
       %Kinvk = L'\(L\k);
       Kinvk =  rm.Kinv*k;
       Kinvdk = rm.Kinv*dk;

       hx1 =   hx -  k'*rm.KinvH;
       dhx1 = dhx - dk'*rm.KinvH;
       c_star  = s - k'*Kinvk + hx1*rm.V_star*hx1';
       dc_star =   -dk'*Kinvk - (k'*Kinvdk)' + dhx1*rm.V_star*hx1' + (hx1*rm.V_star*dhx1')';  
       
       if c_star < 0
           c_star = abs(c_star);
       end
       std =  rm.sigma_hat * sqrt(c_star);
       dstd = rm.sigma_hat/(2*sqrt(c_star)) * dc_star;
    end


end

